@@ -3,7 +3,7 @@
 
 #![deny(unsafe_code)]
 
-use std::pin::Pin;
+use std::pin::{Pin, pin};
 #[cfg(not(feature = "loom"))]
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering::{self, AcqRel, Acquire, Relaxed};
@@ -14,7 +14,7 @@ use loom::sync::atomic::AtomicUsize;
 use crate::opcode::Opcode;
 use crate::pager::SyncResult;
 use crate::sync_primitive::SyncPrimitive;
-use crate::wait_queue::WaitQueue;
+use crate::wait_queue::{Entry, WaitQueue};
 use crate::{Pager, pager};
 
 /// [`Gate`] is a synchronization primitive that blocks tasks from entering a critical section until
@@ -291,11 +291,12 @@ impl Gate {
     /// ```
     #[inline]
     pub async fn enter_async(&self) -> Result<State, Error> {
-        let mut pager = Pager::default();
-        pager.set_entry(WaitQueue::new(self, Opcode::Wait, false));
-        let mut pinned_pager = Pin::new(&mut pager);
+        let mut pinned_pager = pin!(Pager::default());
+        pinned_pager
+            .wait_queue()
+            .construct(self, Opcode::Wait(0), false);
         self.push_wait_queue_entry(&mut pinned_pager, || {});
-        pinned_pager.await
+        pinned_pager.poll_async().await
     }
 
     /// Enters the [`Gate`] asynchronously with a wait callback.
@@ -325,11 +326,12 @@ impl Gate {
     /// ```
     #[inline]
     pub async fn enter_async_with<F: FnOnce()>(&self, wait_callback: F) -> Result<State, Error> {
-        let mut pager = Pager::default();
-        pager.set_entry(WaitQueue::new(self, Opcode::Wait, false));
-        let mut pinned_pager = Pin::new(&mut pager);
+        let mut pinned_pager = pin!(Pager::default());
+        pinned_pager
+            .wait_queue()
+            .construct(self, Opcode::Wait(0), false);
         self.push_wait_queue_entry(&mut pinned_pager, wait_callback);
-        pinned_pager.await
+        pinned_pager.poll_async().await
     }
 
     /// Enters the [`Gate`] synchronously.
@@ -419,9 +421,10 @@ impl Gate {
     /// ```
     #[inline]
     pub fn enter_sync_with<F: FnOnce()>(&self, wait_callback: F) -> Result<State, Error> {
-        let mut pager = Pager::default();
-        pager.set_entry(WaitQueue::new(self, Opcode::Wait, true));
-        let mut pinned_pager = Pin::new(&mut pager);
+        let mut pinned_pager = pin!(Pager::default());
+        pinned_pager
+            .wait_queue()
+            .construct(self, Opcode::Wait(0), true);
         self.push_wait_queue_entry(&mut pinned_pager, wait_callback);
         pinned_pager.poll_sync()
     }
@@ -436,7 +439,7 @@ impl Gate {
     /// # Examples
     ///
     /// ```
-    /// use std::pin::Pin;
+    /// use std::pin::pin;
     /// use std::sync::Arc;
     /// use std::thread;
     ///
@@ -445,8 +448,7 @@ impl Gate {
     ///
     /// let gate = Arc::new(Gate::default());
     ///
-    /// let mut pager = Pager::default();
-    /// let mut pinned_pager = Pin::new(&mut pager);
+    /// let mut pinned_pager = pin!(Pager::default());
     ///
     /// assert!(gate.register_pager(&mut pinned_pager, true));
     /// assert!(!gate.register_pager(&mut pinned_pager, true));
@@ -466,10 +468,10 @@ impl Gate {
         pager: &mut Pin<&mut Pager<'g, Self>>,
         is_sync: bool,
     ) -> bool {
-        if pager.entry().is_some() {
+        if pager.is_registered() {
             return false;
         }
-        pager.set_entry(WaitQueue::new(self, Opcode::Wait, is_sync));
+        pager.wait_queue().construct(self, Opcode::Wait(0), is_sync);
         self.push_wait_queue_entry(pager, || ());
         true
     }
@@ -488,20 +490,17 @@ impl Gate {
         }) {
             Ok(value) | Err(value) => {
                 let mut count = 0;
-                let entry_addr = value & WaitQueue::ADDR_MASK;
                 let prev_state = State::from(value & WaitQueue::DATA_MASK);
                 let next_state = next_state.unwrap_or(prev_state);
                 let result = Self::into_u8(next_state, error);
-                if entry_addr != 0 {
-                    WaitQueue::iter_forward(
-                        WaitQueue::addr_to_ptr(entry_addr),
-                        false,
-                        |entry, _| {
-                            entry.set_result(result);
-                            count += 1;
-                            false
-                        },
-                    );
+                let anchor_ptr = WaitQueue::to_anchor_ptr(value);
+                if !anchor_ptr.is_null() {
+                    let tail_entry_ptr = WaitQueue::to_entry_ptr(anchor_ptr);
+                    Entry::iter_forward(tail_entry_ptr, false, |entry, _| {
+                        entry.set_result(result);
+                        count += 1;
+                        false
+                    });
                 }
                 (prev_state, count)
             }
@@ -515,28 +514,31 @@ impl Gate {
         pager: &mut Pin<&mut Pager<Self>>,
         mut wait_callback: F,
     ) {
-        if let Some(entry) = pager.entry() {
-            let pinned_entry = Pin::new(entry);
-            loop {
-                let state = self.state.load(Acquire);
-                match State::from(state & WaitQueue::DATA_MASK) {
-                    State::Controlled => {
-                        if let Some(returned) =
-                            self.try_push_wait_queue_entry(pinned_entry, state, wait_callback)
-                        {
-                            wait_callback = returned;
-                            continue;
-                        }
-                    }
-                    State::Sealed => {
-                        entry.set_result(Self::into_u8(State::Sealed, Some(Error::Sealed)));
-                    }
-                    State::Open => {
-                        entry.set_result(Self::into_u8(State::Open, None));
+        loop {
+            let state = self.state.load(Acquire);
+            match State::from(state & WaitQueue::DATA_MASK) {
+                State::Controlled => {
+                    if let Some(returned) =
+                        self.try_push_wait_queue_entry(pager.wait_queue(), state, wait_callback)
+                    {
+                        wait_callback = returned;
+                        continue;
                     }
                 }
-                break;
+                State::Sealed => {
+                    pager
+                        .wait_queue()
+                        .entry()
+                        .set_result(Self::into_u8(State::Sealed, Some(Error::Sealed)));
+                }
+                State::Open => {
+                    pager
+                        .wait_queue()
+                        .entry()
+                        .set_result(Self::into_u8(State::Open, None));
+                }
             }
+            break;
         }
     }
 
@@ -569,8 +571,8 @@ impl SyncPrimitive for Gate {
     }
 
     #[inline]
-    fn drop_wait_queue_entry(entry: &WaitQueue) {
-        if entry.try_acknowledge_result().is_none() {
+    fn drop_wait_queue_entry(entry: &Entry) {
+        if entry.try_consume_result().is_none() {
             let this: &Self = entry.sync_primitive_ref();
             this.wake_all(None, Some(Error::SpuriousFailure));
             entry.acknowledge_result_sync();
