@@ -3,7 +3,7 @@
 #![deny(unsafe_code)]
 
 use std::fmt;
-use std::pin::Pin;
+use std::pin::{Pin, pin};
 #[cfg(not(feature = "loom"))]
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering::{self, AcqRel, Acquire, Relaxed, Release};
@@ -16,7 +16,7 @@ use crate::Pager;
 use crate::opcode::Opcode;
 use crate::pager::{self, SyncResult};
 use crate::sync_primitive::SyncPrimitive;
-use crate::wait_queue::{PinnedWaitQueue, WaitQueue};
+use crate::wait_queue::{Entry, PinnedEntry, WaitQueue};
 
 /// [`Lock`] is a low-level locking primitive for both synchronous and asynchronous operations.
 ///
@@ -29,13 +29,23 @@ pub struct Lock {
     state: AtomicUsize,
 }
 
-/// Locking mode, either exclusive or shared.
+/// Operation mode.
 #[derive(Clone, Copy, Debug, Eq, PartialEq)]
 pub enum Mode {
-    /// Exclusive lock.
+    /// Acquires an exclusive lock.
     Exclusive,
-    /// Shared lock.
+    /// Acquires a shared lock.
     Shared,
+    /// Waits for the [`Lock`] to be free or poisoned.
+    ///
+    /// [`Self::WaitExclusive`], [`Self::WaitShared`], [`Lock::try_lock`], and [`Lock::try_share`]
+    /// provide a way to bypass the fair queueing mechanism without spinning.
+    WaitExclusive,
+    /// Waits for a shared lock to be available or the [`Lock`] to be poisoned.
+    ///
+    /// If a [`Self::WaitExclusive`] entry is in front of a [`Self::WaitShared`] entry, the
+    /// [`Self::WaitShared`] entry has to wait until the [`Self::WaitExclusive`] entry is processed.
+    WaitShared,
 }
 
 impl Lock {
@@ -156,9 +166,7 @@ impl Lock {
     /// ```
     #[inline]
     pub async fn lock_async(&self) -> bool {
-        let async_wait = WaitQueue::new(self, Opcode::Exclusive, false);
-        self.acquire_async_with_internal::<_, true>(&async_wait, || {})
-            .await
+        self.acquire_async_with_internal::<_, true>(|| {}).await
     }
 
     /// Acquires an exclusive lock asynchronously with a wait callback.
@@ -182,8 +190,7 @@ impl Lock {
     /// ```
     #[inline]
     pub async fn lock_async_with<F: FnOnce()>(&self, begin_wait: F) -> bool {
-        let async_wait = WaitQueue::new(self, Opcode::Exclusive, false);
-        self.acquire_async_with_internal::<_, true>(&async_wait, begin_wait)
+        self.acquire_async_with_internal::<_, true>(begin_wait)
             .await
     }
 
@@ -288,9 +295,7 @@ impl Lock {
     /// ```
     #[inline]
     pub async fn share_async(&self) -> bool {
-        let async_wait = WaitQueue::new(self, Opcode::Shared, false);
-        self.acquire_async_with_internal::<_, false>(&async_wait, || ())
-            .await
+        self.acquire_async_with_internal::<_, false>(|| ()).await
     }
 
     /// Acquires a shared lock asynchronously with a wait callback.
@@ -314,8 +319,7 @@ impl Lock {
     /// ```
     #[inline]
     pub async fn share_async_with<F: FnOnce()>(&self, begin_wait: F) -> bool {
-        let async_wait = WaitQueue::new(self, Opcode::Shared, false);
-        self.acquire_async_with_internal::<_, false>(&async_wait, begin_wait)
+        self.acquire_async_with_internal::<_, false>(begin_wait)
             .await
     }
 
@@ -411,15 +415,14 @@ impl Lock {
     /// # Examples
     ///
     /// ```
-    /// use std::pin::Pin;
+    /// use std::pin::pin;
     ///
     /// use saa::{Lock, Pager};
     /// use saa::lock::Mode;
     ///
     /// let lock = Lock::default();
     ///
-    /// let mut pager = Pager::default();
-    /// let mut pinned_pager = Pin::new(&mut pager);
+    /// let mut pinned_pager = pin!(Pager::default());
     ///
     /// assert!(lock.register_pager(&mut pinned_pager, Mode::Exclusive, true));
     /// assert!(!lock.register_pager(&mut pinned_pager, Mode::Exclusive, true));
@@ -433,30 +436,45 @@ impl Lock {
         mode: Mode,
         is_sync: bool,
     ) -> bool {
-        if pager.entry().is_some() {
+        if pager.is_registered() {
             return false;
         }
         let opcode = match mode {
             Mode::Exclusive => Opcode::Exclusive,
             Mode::Shared => Opcode::Shared,
+            Mode::WaitExclusive => Opcode::Wait(u8::try_from(WaitQueue::DATA_MASK).unwrap_or(0)),
+            Mode::WaitShared => Opcode::Wait(1),
         };
-        pager.set_entry(WaitQueue::new(self, opcode, is_sync));
+
+        pager.wait_queue().construct(self, opcode, is_sync);
+
         loop {
-            let Some(entry) = pager.entry() else {
-                continue;
-            };
             let (result, state) = match mode {
                 Mode::Exclusive => self.try_lock_internal(),
                 Mode::Shared => self.try_share_internal(),
+                Mode::WaitExclusive | Mode::WaitShared => {
+                    let state = self.state.load(Acquire);
+                    let result = if state == usize::from(Self::POISONED) {
+                        Self::POISONED
+                    } else if (mode == Mode::WaitExclusive && (state & WaitQueue::DATA_MASK) == 0)
+                        || (mode == Mode::WaitShared
+                            && (state & WaitQueue::DATA_MASK) < Self::MAX_SHARED_OWNERS)
+                    {
+                        // If an available lock is available, return immediately.
+                        Self::ACQUIRED
+                    } else {
+                        Self::NOT_ACQUIRED
+                    };
+                    (result, state)
+                }
             };
             if result == Self::ACQUIRED || result == Self::POISONED {
-                entry.set_result(result);
+                pager.wait_queue().entry().set_result(result);
                 break;
             }
 
-            let pinned_entry = Pin::new(entry);
             if self
-                .try_push_wait_queue_entry(pinned_entry, state, || ())
+                .try_push_wait_queue_entry(pager.wait_queue(), state, || ())
                 .is_none()
             {
                 break;
@@ -643,10 +661,8 @@ impl Lock {
     #[inline]
     async fn acquire_async_with_internal<F: FnOnce(), const EXCLUSIVE: bool>(
         &self,
-        wait_queue: &WaitQueue,
         mut begin_wait: F,
     ) -> bool {
-        let pinned_entry = PinnedWaitQueue(Pin::new(wait_queue));
         loop {
             let (mut result, state) = if EXCLUSIVE {
                 self.try_lock_internal()
@@ -661,14 +677,21 @@ impl Lock {
             debug_assert_eq!(result, Self::NOT_ACQUIRED);
             debug_assert!(state & WaitQueue::ADDR_MASK != 0 || state & WaitQueue::DATA_MASK != 0);
 
+            let mode = if EXCLUSIVE {
+                Opcode::Exclusive
+            } else {
+                Opcode::Shared
+            };
+            let async_wait = pin!(WaitQueue::default());
+            async_wait.as_ref().construct(self, mode, false);
             if let Some(returned) =
-                self.try_push_wait_queue_entry(pinned_entry.0, state, begin_wait)
+                self.try_push_wait_queue_entry(async_wait.as_ref(), state, begin_wait)
             {
                 begin_wait = returned;
                 continue;
             }
 
-            result = pinned_entry.await;
+            result = PinnedEntry(Pin::new(async_wait.entry())).await;
             debug_assert!(result == Self::ACQUIRED || result == Self::POISONED);
             return result == Self::ACQUIRED;
         }
@@ -729,16 +752,13 @@ impl Lock {
                     // woke up the current lock owner has finished processing the wait queue is
                     // prevented by the wait queue processing method itself; `model.rs` proves it.
                     debug_assert_eq!(prev_state & WaitQueue::LOCKED_FLAG, 0);
-                    let entry_addr = prev_state & WaitQueue::ADDR_MASK;
-                    if entry_addr != 0 {
-                        WaitQueue::iter_forward(
-                            WaitQueue::addr_to_ptr(entry_addr),
-                            false,
-                            |entry, _| {
-                                entry.set_result(Self::POISONED);
-                                false
-                            },
-                        );
+                    let anchor_ptr = WaitQueue::to_anchor_ptr(prev_state);
+                    if !anchor_ptr.is_null() {
+                        let tail_entry_ptr = WaitQueue::to_entry_ptr(anchor_ptr);
+                        Entry::iter_forward(tail_entry_ptr, false, |entry, _| {
+                            entry.set_result(Self::POISONED);
+                            false
+                        });
                     }
 
                     return true;
@@ -781,7 +801,7 @@ impl SyncPrimitive for Lock {
     }
 
     #[inline]
-    fn drop_wait_queue_entry(entry: &WaitQueue) {
+    fn drop_wait_queue_entry(entry: &Entry) {
         Self::force_remove_wait_queue_entry(entry);
     }
 }
