@@ -4,7 +4,7 @@
 #![deny(unsafe_code)]
 
 use std::fmt;
-use std::pin::Pin;
+use std::pin::{Pin, pin};
 #[cfg(not(feature = "loom"))]
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering::{self, Acquire, Relaxed, Release};
@@ -16,7 +16,7 @@ use crate::Pager;
 use crate::opcode::Opcode;
 use crate::pager::{self, SyncResult};
 use crate::sync_primitive::SyncPrimitive;
-use crate::wait_queue::{PinnedWaitQueue, WaitQueue};
+use crate::wait_queue::{Entry, PinnedEntry, WaitQueue};
 
 /// [`Semaphore`] is a synchronization primitive that allows a fixed number of threads to access a
 /// resource concurrently.
@@ -138,9 +138,7 @@ impl Semaphore {
     /// ```
     #[inline]
     pub async fn acquire_async(&self) {
-        let async_wait = WaitQueue::new(self, Opcode::Semaphore(1), false);
-        self.acquire_async_with_internal(&async_wait, 1, || {})
-            .await;
+        self.acquire_async_with_internal(1, || {}).await;
     }
 
     /// Gets a permit from the semaphore asynchronously with a wait callback.
@@ -164,9 +162,7 @@ impl Semaphore {
     /// ```
     #[inline]
     pub async fn acquire_async_with<F: FnOnce()>(&self, begin_wait: F) {
-        let async_wait = WaitQueue::new(self, Opcode::Semaphore(1), false);
-        self.acquire_async_with_internal(&async_wait, 1, begin_wait)
-            .await;
+        self.acquire_async_with_internal(1, begin_wait).await;
     }
 
     /// Gets a permit from the semaphore synchronously.
@@ -247,10 +243,7 @@ impl Semaphore {
     /// ```
     #[inline]
     pub async fn acquire_many_async(&self, count: usize) -> bool {
-        #[allow(clippy::cast_possible_truncation)]
-        let async_wait = WaitQueue::new(self, Opcode::Semaphore(count as u8), false);
-        self.acquire_async_with_internal(&async_wait, count, || {})
-            .await
+        self.acquire_async_with_internal(count, || {}).await
     }
 
     /// Gets multiple permits from the semaphore asynchronously with a wait callback.
@@ -275,10 +268,7 @@ impl Semaphore {
     /// ```
     #[inline]
     pub async fn acquire_many_async_with<F: FnOnce()>(&self, count: usize, begin_wait: F) -> bool {
-        #[allow(clippy::cast_possible_truncation)]
-        let async_wait = WaitQueue::new(self, Opcode::Semaphore(count as u8), false);
-        self.acquire_async_with_internal(&async_wait, count, begin_wait)
-            .await
+        self.acquire_async_with_internal(count, begin_wait).await
     }
 
     /// Gets multiple permits from the semaphore synchronously.
@@ -379,14 +369,13 @@ impl Semaphore {
     /// # Examples
     ///
     /// ```
-    /// use std::pin::Pin;
+    /// use std::pin::pin;
     ///
     /// use saa::{Pager, Semaphore};
     ///
     /// let semaphore = Semaphore::default();
     ///
-    /// let mut pager = Pager::default();
-    /// let mut pinned_pager = Pin::new(&mut pager);
+    /// let mut pinned_pager = pin!(Pager::default());
     ///
     /// assert!(semaphore.register_pager(&mut pinned_pager, 1, true));
     /// assert!(!semaphore.register_pager(&mut pinned_pager, 1, true));
@@ -400,27 +389,26 @@ impl Semaphore {
         count: usize,
         is_sync: bool,
     ) -> bool {
-        if pager.entry().is_some() || count > Self::MAX_PERMITS {
+        if count > Self::MAX_PERMITS || pager.is_registered() {
             return false;
         }
         let Ok(count) = u8::try_from(count) else {
             return false;
         };
 
-        pager.set_entry(WaitQueue::new(self, Opcode::Semaphore(count), is_sync));
+        pager
+            .wait_queue()
+            .construct(self, Opcode::Semaphore(count), is_sync);
+
         loop {
-            let Some(entry) = pager.entry() else {
-                continue;
-            };
             let (result, state) = self.try_acquire_internal(count);
             if result {
-                entry.set_result(0);
+                pager.wait_queue().entry().set_result(0);
                 break;
             }
 
-            let pinned_entry = Pin::new(entry);
             if self
-                .try_push_wait_queue_entry(pinned_entry, state, || ())
+                .try_push_wait_queue_entry(pager.wait_queue(), state, || ())
                 .is_none()
             {
                 break;
@@ -491,7 +479,6 @@ impl Semaphore {
     #[inline]
     async fn acquire_async_with_internal<F: FnOnce()>(
         &self,
-        wait_queue: &WaitQueue,
         count: usize,
         mut begin_wait: F,
     ) -> bool {
@@ -501,7 +488,6 @@ impl Semaphore {
         let Ok(count) = u8::try_from(count) else {
             return false;
         };
-        let pinned_entry = PinnedWaitQueue(Pin::new(wait_queue));
         loop {
             let (result, state) = self.try_acquire_internal(count);
             if result {
@@ -509,14 +495,18 @@ impl Semaphore {
             }
             debug_assert!(state & WaitQueue::ADDR_MASK != 0 || state & WaitQueue::DATA_MASK != 0);
 
+            let async_wait = pin!(WaitQueue::default());
+            async_wait
+                .as_ref()
+                .construct(self, Opcode::Semaphore(count), false);
             if let Some(returned) =
-                self.try_push_wait_queue_entry(pinned_entry.0, state, begin_wait)
+                self.try_push_wait_queue_entry(async_wait.as_ref(), state, begin_wait)
             {
                 begin_wait = returned;
                 continue;
             }
 
-            pinned_entry.await;
+            PinnedEntry(Pin::new(async_wait.entry())).await;
             return true;
         }
     }
@@ -571,7 +561,7 @@ impl SyncPrimitive for Semaphore {
     }
 
     #[inline]
-    fn drop_wait_queue_entry(entry: &WaitQueue) {
+    fn drop_wait_queue_entry(entry: &Entry) {
         Self::force_remove_wait_queue_entry(entry);
     }
 }
