@@ -1,63 +1,50 @@
-use std::pin::Pin;
+use std::pin::pin;
 use std::sync::atomic::AtomicUsize;
 use std::sync::atomic::Ordering::Relaxed;
 use std::sync::{Arc, Barrier};
 use std::thread;
 use std::time::Duration;
 
-use crate::wait_queue::WaitQueue;
 use crate::{Gate, Lock, Pager, Semaphore, gate, lock};
 
 #[test]
 fn future_size() {
+    let limit = 196;
     let lock = Lock::default();
 
     let lock_fut = &lock.lock_async();
-    assert_eq!(size_of_val(lock_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(lock_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(lock_fut) < limit);
 
     let lock_with_fut = &lock.lock_async_with(|| {});
-    assert_eq!(size_of_val(lock_with_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(lock_with_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(lock_with_fut) < limit);
 
     let share_fut = &lock.share_async();
-    assert_eq!(size_of_val(share_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(share_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(share_fut) < limit);
 
     let share_with_fut = &lock.share_async_with(|| {});
-    assert_eq!(size_of_val(share_with_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(share_with_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(share_with_fut) < limit);
 
     let semaphore = Semaphore::default();
 
     let acquire_fut = &semaphore.acquire_async();
-    assert_eq!(size_of_val(acquire_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(acquire_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(acquire_fut) < limit);
 
     let acquire_with_fut = &semaphore.acquire_async_with(|| {});
-    assert_eq!(size_of_val(acquire_with_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(acquire_with_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(acquire_with_fut) < limit);
 
     let acquire_many_fut = &semaphore.acquire_many_async(1);
-    assert_eq!(size_of_val(acquire_many_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(acquire_many_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(acquire_many_fut) < limit);
 
     let acquire_many_with_fut = &semaphore.acquire_many_async_with(1, || {});
-    assert_eq!(
-        size_of_val(acquire_many_with_fut),
-        size_of::<WaitQueue>() * 2
-    );
-    assert_eq!(align_of_val(acquire_many_with_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(acquire_many_with_fut) < limit);
 
     let gate = Gate::default();
 
     let enter_fut = &gate.enter_async();
-    assert_eq!(size_of_val(enter_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(enter_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(enter_fut) < limit);
 
     let enter_with_fut = &gate.enter_async_with(|| {});
-    assert_eq!(size_of_val(enter_with_fut), size_of::<WaitQueue>() * 2);
-    assert_eq!(align_of_val(enter_with_fut), align_of::<WaitQueue>());
+    assert!(size_of_val(enter_with_fut) < limit);
 }
 
 #[cfg_attr(miri, ignore = "Tokio is not compatible with Miri")]
@@ -462,11 +449,10 @@ async fn lock_pager() {
         let lock = lock.clone();
         if i % 2 == 0 {
             tasks.push(tokio::spawn(async move {
-                let mut pager = Pager::default();
+                let mut pinned_pager = pin!(Pager::default());
                 for _ in 0..num_iters {
-                    let mut pinned_pager = Pin::new(&mut pager);
                     lock.register_pager(&mut pinned_pager, lock::Mode::Shared, false);
-                    match pinned_pager.await {
+                    match pinned_pager.poll_async().await {
                         Ok(true) => {
                             check.fetch_add(1, Relaxed);
                             lock.release_share();
@@ -477,9 +463,8 @@ async fn lock_pager() {
             }));
         } else {
             threads.push(thread::spawn(move || {
-                let mut pager = Pager::default();
+                let mut pinned_pager = pin!(Pager::default());
                 for _ in 0..num_iters {
-                    let mut pinned_pager = Pin::new(&mut pager);
                     lock.register_pager(&mut pinned_pager, lock::Mode::Exclusive, true);
                     match pinned_pager.poll_sync() {
                         Ok(true) => {
@@ -503,6 +488,84 @@ async fn lock_pager() {
     assert_eq!(check.load(Relaxed), num_iters * num_tasks);
 }
 
+#[test]
+fn lock_pager_wait() {
+    let lock = Lock::default();
+
+    assert!(lock.try_lock());
+    let mut pinned_pager_1 = pin!(Pager::default());
+    let mut pinned_pager_2 = pin!(Pager::default());
+    let mut pinned_pager_3 = pin!(Pager::default());
+    assert!(lock.register_pager(&mut pinned_pager_1, lock::Mode::Shared, false));
+    assert!(lock.register_pager(&mut pinned_pager_2, lock::Mode::WaitExclusive, false));
+    assert!(lock.register_pager(&mut pinned_pager_3, lock::Mode::Exclusive, false));
+    assert!(lock.release_lock());
+    assert_eq!(pinned_pager_1.try_poll(), Ok(true));
+    assert!(pinned_pager_2.try_poll().is_err());
+    assert!(pinned_pager_3.try_poll().is_err());
+
+    assert!(lock.release_share());
+    assert_eq!(pinned_pager_2.try_poll(), Ok(true));
+    assert_eq!(pinned_pager_3.try_poll(), Ok(true));
+
+    assert!(lock.register_pager(&mut pinned_pager_2, lock::Mode::WaitShared, false));
+    assert!(pinned_pager_2.try_poll().is_err());
+    assert!(lock.poison_lock());
+    assert_eq!(pinned_pager_2.try_poll(), Ok(false));
+}
+
+#[cfg_attr(miri, ignore = "Tokio is not compatible with Miri")]
+#[tokio::test(flavor = "multi_thread", worker_threads = 16)]
+async fn lock_pager_wait_parallel() {
+    let num_tasks = 8;
+    let num_iters = 64;
+    let lock = Arc::new(Lock::default());
+
+    let mut threads = Vec::new();
+    let mut tasks = Vec::new();
+    for i in 0..num_tasks {
+        let lock = lock.clone();
+        if i % 4 == 0 {
+            tasks.push(tokio::spawn(async move {
+                let mut pinned_pager = pin!(Pager::default());
+                for _ in 0..num_iters {
+                    let mode = if num_iters % 7 == 0 {
+                        lock::Mode::Exclusive
+                    } else {
+                        lock::Mode::Shared
+                    };
+                    lock.register_pager(&mut pinned_pager, mode, false);
+                    match pinned_pager.poll_async().await {
+                        Ok(true) => {
+                            if mode == lock::Mode::Exclusive {
+                                lock.release_lock();
+                            } else {
+                                lock.release_share();
+                            }
+                        }
+                        _ => unreachable!(),
+                    }
+                }
+            }));
+        } else {
+            threads.push(thread::spawn(move || {
+                let mut pinned_pager = pin!(Pager::default());
+                for _ in 0..num_iters {
+                    lock.register_pager(&mut pinned_pager, lock::Mode::WaitShared, true);
+                    assert!(pinned_pager.poll_sync().is_ok());
+                }
+            }));
+        }
+    }
+
+    for thread in threads {
+        thread.join().unwrap();
+    }
+    for task in tasks {
+        task.await.unwrap();
+    }
+}
+
 #[cfg_attr(miri, ignore = "Tokio is not compatible with Miri")]
 #[tokio::test(flavor = "multi_thread", worker_threads = 16)]
 async fn semaphore_async() {
@@ -684,12 +747,11 @@ async fn semaphore_pager() {
         let semaphore = semaphore.clone();
         if i % 2 == 0 {
             tasks.push(tokio::spawn(async move {
-                let mut pager = Pager::default();
+                let mut pinned_pager = pin!(Pager::default());
                 for _ in 0..num_iters {
-                    let mut pinned_pager = Pin::new(&mut pager);
                     assert!(!semaphore.register_pager(&mut pinned_pager, usize::MAX, false));
                     semaphore.register_pager(&mut pinned_pager, Semaphore::MAX_PERMITS, false);
-                    match pinned_pager.await {
+                    match pinned_pager.poll_async().await {
                         Ok(()) => {
                             check.fetch_add(1, Relaxed);
                             semaphore.release_many(Semaphore::MAX_PERMITS);
@@ -700,9 +762,8 @@ async fn semaphore_pager() {
             }));
         } else {
             threads.push(thread::spawn(move || {
-                let mut pager = Pager::default();
+                let mut pinned_pager = pin!(Pager::default());
                 for _ in 0..num_iters {
-                    let mut pinned_pager = Pin::new(&mut pager);
                     assert!(!semaphore.register_pager(&mut pinned_pager, usize::MAX, true));
                     semaphore.register_pager(&mut pinned_pager, Semaphore::MAX_PERMITS, true);
                     match pinned_pager.poll_sync() {
@@ -814,11 +875,10 @@ async fn gate_pager() {
         let gate = gate.clone();
         if i % 2 == 0 {
             tasks.push(tokio::spawn(async move {
-                let mut pager = Pager::default();
+                let mut pinned_pager = pin!(Pager::default());
                 for _ in 0..num_iters {
-                    let mut pinned_pager = Pin::new(&mut pager);
                     gate.register_pager(&mut pinned_pager, false);
-                    match pinned_pager.await {
+                    match pinned_pager.poll_async().await {
                         Ok(_) => {
                             check.fetch_add(1, Relaxed);
                         }
@@ -828,9 +888,8 @@ async fn gate_pager() {
             }));
         } else {
             threads.push(thread::spawn(move || {
-                let mut pager = Pager::default();
+                let mut pinned_pager = pin!(Pager::default());
                 for _ in 0..num_iters {
-                    let mut pinned_pager = Pin::new(&mut pager);
                     gate.register_pager(&mut pinned_pager, true);
                     match pinned_pager.poll_sync() {
                         Ok(_) => {
@@ -880,8 +939,7 @@ fn pager_drop_implicit() {
     }
 
     for is_sync in [false, true] {
-        let mut pager = Pager::default();
-        let mut pinned_pager = Pin::new(&mut pager);
+        let mut pinned_pager = pin!(Pager::default());
         lock.register_pager(&mut pinned_pager, lock::Mode::Exclusive, is_sync);
     }
 
@@ -906,8 +964,7 @@ fn pager_drop_explicit() {
     });
 
     {
-        let mut pager = Pager::default();
-        let mut pinned_pager = Pin::new(&mut pager);
+        let mut pinned_pager = pin!(Pager::default());
         assert!(semaphore.register_pager(&mut pinned_pager, 11, false));
     }
 
@@ -951,8 +1008,7 @@ async fn lock_chaos() {
                 for j in 0..num_iters {
                     assert!(!lock.is_poisoned(Relaxed));
                     if j % 7 == 3 {
-                        let mut pager = Pager::default();
-                        let mut pinned_pager = Pin::new(&mut pager);
+                        let mut pinned_pager = pin!(Pager::default());
                         lock.register_pager(&mut pinned_pager, lock::Mode::Exclusive, j % 2 == 0);
                     } else if j % 11 == 0 {
                         assert!(lock.lock_sync());
@@ -1008,8 +1064,7 @@ async fn semaphore_chaos() {
             threads.push(thread::spawn(move || {
                 for j in 0..num_iters {
                     if j % 7 == 1 {
-                        let mut pager = Pager::default();
-                        let mut pinned_pager = Pin::new(&mut pager);
+                        let mut pinned_pager = pin!(Pager::default());
                         assert!(semaphore.register_pager(&mut pinned_pager, 27, j % 2 == 0));
                     } else {
                         assert!(semaphore.acquire_many_sync(i + 1));
@@ -1063,14 +1118,16 @@ async fn gate_chaos() {
             threads.push(thread::spawn(move || {
                 for j in 0..num_iters {
                     if j % 7 == 5 {
-                        let mut pager = Pager::default();
-                        let mut pinned_pager = Pin::new(&mut pager);
+                        let mut pinned_pager = pin!(Pager::default());
                         if i % 3 == 0 {
                             assert!(gate.register_pager(&mut pinned_pager, false));
                         } else {
                             assert!(gate.register_pager(&mut pinned_pager, true));
-                            if let Ok(state) = pinned_pager.try_poll() {
-                                assert_eq!(pinned_pager.poll_sync(), Ok(state));
+                            if pinned_pager.try_poll().is_ok() {
+                                assert_eq!(
+                                    pinned_pager.poll_sync(),
+                                    Err(gate::Error::NotRegistered)
+                                );
                             }
                         }
                     } else {
